/*
   "disk to file" program.

   Written 1/26/98 by WAC.
   Modified 4/15/98 by WAC for 360K / 720K floppy compatibility
   Modified 2/13/00 by WAC to make this a more general utility -- add
                           comment field & directory listing
   Modified 10/8/00 by WAC to add 160K / 180K / 320K floppies
*/

#include "dskfile2.h"
#include "crc16.h"
#include "keydefs.h"

#define MIN(a,b)  (((a) < (b)) ? (a) : (b))
#define DEPTH 25      /* directory depth +1 that program will handle */

struct hdr_format header;

BYTE diskbuf[DISKBUF_SIZE];
BYTE outbuf[OUTBUF_SIZE];
WORD outbuf_count;  /* since OUTBUF can be of variable size, need a size count */

WORD spt;
WORD heads;
WORD cyls;
WORD drive;  /* 0=A, 1=B */

int handle;

BYTE filename[128];
WORD updatecrc();  /* pre-declare type */

#define MAX_COMMENT_LINES 24
#define MAX_COMMENT_LINE_LEN 78
struct comment_entry
   {
   struct comment_line
      {
      WORD linelen;
      BYTE buf[MAX_COMMENT_LINE_LEN+2];
      } line[MAX_COMMENT_LINES];
   } comments;

#define MODE_INSERT 0
#define MODE_OVERSTRIKE 1

BYTE eraseline[80] =
"                                                                              ";

BYTE fname[20];
BYTE wildname[20] = "*.*";
BYTE saved_drive;
BYTE dirname[128];
BYTE currentdir[DEPTH][128];
BYTE path[128];

BYTE dta[DEPTH][64];
WORD dta_ofs,dta_seg;
WORD dirlevel = 1;
WORD max_dirlevel = 1;
WORD start;

#define ZIP_LOCALHDR     0x04034b50L
#define ZIP_CENTRALHDR   0x02014b50L
#define ZIP_HDRSIZE      0x1e
#define ZIP_HAVE_DATA_DESC   0x08  /* bit 3 */
struct ziplocalhdr
   {
   DWORD signature;
   WORD  extract_ver;
   WORD  bitflag;
   WORD  compr_method;
   WORD  file_time;
   WORD  file_date;
   DWORD crc32;
   DWORD comprsize;
   DWORD realsize;
   WORD  namelen;
   WORD  extra_field_len;
   BYTE  filename[80];
   } zip_localhdr;

#define ARC_SIGNATURE 0x1a
#define ARC_EOF 0x00
#define ARC_HDRSIZE 0x1d
struct archeader
   {
   BYTE signature;    /* 0x1a if ARC file */
   BYTE stowage;      /* compression method, or 0 (ARC_EOF) at EOF */
   BYTE filename[13]; /* 8.3 name, zero-byte terminated */
   DWORD comprsize;   /* compressed size */
   WORD file_date;    /* date in MSDOS directory format */
   WORD file_time;    /* time in MSDOS directory format */
   WORD crc16;        /* crc-16 of file */
   DWORD realsize;    /* uncompressed file size */
   } arc_header;

/* NOTE: ARJ_MAX_HEADER_SIZE defined in DSKFILE2.H, so that OUTBUF can
 *       can be used as a temporary buf for ARJ header processing
 */
#define ARJ_SIGNATURE 0xEA60
#define ARJ_EOF 0x0000
#define ARJ_EXTFILE_FLAG 0x08  /* bit 3 */
#define ARJ_EXTFILE_SIZE 4  /* bytes to add if ARJ_EXTFILE_FLAG set */
#define ARJ_CRC_SIZE 4
struct arjheader
   {
   WORD signature;   /* 0xEA60 if ARJ file */
   WORD basic_header_size;  /* size to end of comment, so
                  ofs to end of basic header + basic_header_size + CRC len +
                  sizeof(extended header len) + extended header len +
                  extended header CRC len + data len = next header <whew!>
                  Also, basic_header_size == 0  ->  end of .ARJ */
   BYTE first_header_size;  /* size up to & including "extra data" */
   BYTE archiver_version;
   BYTE min_extract_version;
   BYTE host_os;
   BYTE arj_flags;  /* 0x08 == EXTFILE_FLAG, so extra_data exists */
   BYTE method;     /* compression method */
   BYTE file_type;  /* 0 = binary, 1 = 7-bit text, 2 = main ARJ header,
                       3 = directory, 4 = volume label */
   BYTE reserved;
   WORD file_time;    /* time in MSDOS directory format */
   WORD file_date;    /* date in MSDOS directory format */
   DWORD comprsize;   /* compressed size */
   DWORD realsize;    /* uncompressed file size */
   DWORD crc32;       /* original (uncompressed) file CRC */
   WORD filespec_pos; /* position of name in stored name/path */
   WORD fileaccess_mode;
   WORD host_data;    /* not currently used in ARJ header (April 1993 docs) */
          /*
           * <DWORD extra_data if ARJ_EXTFILE_FLAG set>
           * <?? bytes name, EOS <zero-byte> terminated>
           * <?? bytes comment, EOS <zero-byte> terminated>
           * <DWORD basic header CRC32>
           * <WORD 1st extended header size (0 if none)>
           * <?? bytes 1st extended header if size != 0>
           * <DWORD 1st extended header CRC32 if size != 0>
           * <?? bytes compressed file data>
           */
   } arj_header;

#define DB_SEPERATOR '\014'  /* FormFeed char */
#define DB_FIELD_SEP '\024'  /* ^T, visually the '' char */
WORD update_database = TRUE;
int db_handle, db_handle2;
BYTE db_field_str[4] =
    { DB_FIELD_SEP, '\015', '\012' };

WORD do_extra_tracks = FALSE;

DWORD lseek();

main(argc,argv)
   WORD argc;
   BYTE *argv[];
   {
   WORD i, j, h, c;
   DWORD db_pos;

   for (i = 1; i < argc; i++)
      {
      if (argv[i][0] == '-')
         {
         if (toupper(argv[i][1]) == 'D')
            update_database = FALSE;
         else
            {
            if (toupper(argv[i][1]) == 'E')
               do_extra_tracks = TRUE;
            else
               help();
            }
         for (j = i; j < argc-1; j++)
            argv[j] = argv[j+1];
         argc--;
         i--;
         }
      }
   if ((argc != 3) || (strlen(argv[1]) > 2))
      help();
   if ((strlen(argv[1]) == 2) && (argv[1][1] != ':'))
      help();
   drive = toupper(argv[1][0]);
   if ((drive != 'A') && (drive != 'B'))
      help();
   if (update_database == FALSE)
      {
      printf("Do you want to create a disk image without putting it\n");
      printf("in the database (Y/N)? ");
      gets(diskbuf);
      puts("\n");
      if (toupper(diskbuf[0]) != 'Y')
         exit(1);
      }
   drive -= 'A';  /* drive letter to BIOS number */
   strcpy(filename,argv[2]);  /* get a global copy */
   if ((handle = open(filename,MODE_READ)) != FILE_ERROR)
      {
      close(handle);
      printf("The file %s already exists.\n",filename);
      puts("Do you want to overwrite it (Y/N)? ");
      gets(diskbuf);
      puts("\n");
      if (toupper(diskbuf[0]) != 'Y')
         exit(1);
      }
   if ((handle = creat(filename)) == FILE_ERROR)
      {
      puts("ERROR:  Cannot create output file!\n");
      exit(2);
      }
   if (!ident_disk())  /* rest of header filled in by this routine */
      {
      close(handle);
      unlink(filename);
      exit(2);
      }
   if ((db_handle = creat("C:\\DISKLIST.TMP")) == -1)
      {
      puts("Error: unable to update DISKLIST.DB. Continue without updating? (Y/N)");
      if (toupper(ci()) != 'Y')
         exit(2);
      update_database = FALSE;
      }
   get_comments();
   for (i = 0; i < MAX_COMMENT_LINES; i++)
      if (comments.line[i].linelen)
         break;  /* if any non-blank comment lines, i < MAX_COMMENT_LINES */
   if (i < MAX_COMMENT_LINES)
      header.extrasflag |= EFLAG_COMMENT;
   if (write(handle,&header,sizeof(header)) != sizeof(header))
      {
      close(handle);
      unlink(filename);
      if (update_database)
         {
         close(db_handle);
         unlink("C:\\DISKLIST.TMP");
         }
      puts("Error writing to output file.  Disk full/write protected?\n");
      exit(2);
      }
   for (c = 0; c < cyls; c++)
      {
      for (h = 0; h < heads; h++)
         {
         build_outbuf(h,c);
         write_outbuf();   /* this does not return if an error happens */
         }
      }
   if (header.head0extra1)
      {
      build_outbuf(0,cyls);
      write_outbuf();   /* this does not return if an error happens */
      }
   if (header.head1extra1)
      {
      build_outbuf(1,cyls);
      write_outbuf();   /* this does not return if an error happens */
      }
   if (header.head0extra2)
      {
      build_outbuf(0,cyls+1);
      write_outbuf();   /* this does not return if an error happens */
      }
   if (header.head1extra2)
      {
      build_outbuf(1,cyls+1);
      write_outbuf();   /* this does not return if an error happens */
      }
   if (update_database)
      {
      sprintf(outbuf,"%s is an image of a ",filename);
      if (spt == SPT_320160)
         {
         if (heads == 2)
            strcat(outbuf,"320K disk.\015\012");
         else
            strcat(outbuf,"160K disk.\015\012");
         }
      else
         {
         if (spt == SPT_720360)
            {
            if (cyls == MAX_CYL)
               strcat(outbuf,"720K disk.\015\012");
            else
               {
               if (heads == 2)
                  strcat(outbuf,"360K disk.\015\012");
               else
                  strcat(outbuf,"180K disk.\015\012");
               }
            }
         else
            {
            if (spt == SPT_12)
               strcat(outbuf,"1.2M disk.\015\012");
            else
               strcat(outbuf,"1.44M disk.\015\012");
            }
         }
      if ((write(db_handle,outbuf,strlen(outbuf)) != strlen(outbuf))
              || (write(db_handle,db_field_str,3) != 3))
         {
         puts("Error -- unable to write to DISKLIST.DB!  Disk full?\n");
         close(handle);
         unlink(filename);
         close(db_handle);
         unlink("C:\\DISKLIST.TMP");
         exit(2);
         }
      }
   if (!add_comments())
      {
      close(handle);
      unlink(filename);
      if (update_database)
         {
         close(db_handle);
         unlink("C:\\DISKLIST.TMP");
         }
      puts("\nError writing to output file.  Disk full/write protected?\n");
      exit(2);
      }
   if (!add_dirlist())
      {
      close(handle);
      unlink(filename);
      if (update_database)
         {
         close(db_handle);
         unlink("C:\\DISKLIST.TMP");
         }
      puts("\nError writing to output file.  Disk full/write protected?\n");
      exit(2);
      }
   if (header.extrasflag & EFLAG_DIR)  /* if we have DIR info */
      {
      lseek(handle,0L,0);
      if (write(handle,&header,sizeof(header)) != sizeof(header))
         {
         close(handle);
         unlink(filename);
         if (update_database)
            {
            close(db_handle);
            unlink("C:\\DISKLIST.TMP");
            }
         puts("Error writing to output file.  Disk full/write protected?\n");
         exit(2);
         }
      }
   if (close(handle) == FILE_ERROR)
      {
      unlink(filename);
      if (update_database)
         {
         close(db_handle);
         unlink("C:\\DISKLIST.TMP");
         }
      puts("\nError writing to output file.  Disk full/write protected?\n");
      exit(2);
      }
   if (update_database)
     {
     lseek(db_handle,0L,0);
     if ((db_handle2 = open("C:\\DISKLIST.DB",2)) == -1)
        {
        if ((db_handle2 = creat("C:\\DISKLIST.DB")) == -1)
           {
           close(db_handle);
           unlink("C:\\DISKLIST.TMP");
           puts("Error -- unable to update C:\\DISKLIST.DB!  Disk full?\n");
           exit(2);
           }
        }
     db_pos = lseek(db_handle2,0L,2);  /* lseek EOF */
     while (i = read(db_handle,diskbuf,DISKBUF_SIZE))
        {
        if (write(db_handle2,diskbuf,i) != i)
           {
           close(db_handle);
           close(db_handle2);
           unlink("C:\\DISKLIST.TMP");
           truncate("C:\\DISKLIST.DB",db_pos);
           puts("Error updating C:\\DISKLIST.DB!  Disk full?\n");
           exit(2);
           }
        }
     close(db_handle);
     unlink("C:\\DISKLIST.TMP");
     diskbuf[0] = DB_SEPERATOR;  /* FormFeed char = seperator */
     diskbuf[1] = '\015';
     diskbuf[2] = '\012';
     if (write(db_handle2,diskbuf,3) != 3)
        {
        close(db_handle2);
        truncate("C:\\DISKLIST.DB",db_pos);
        puts("Error updating C:\\DISKLIST.DB!  Disk full?\n");
        exit(2);
        }
     close(db_handle2);
     }
   exit(0);
   }

help()
   {
   puts("The format is:\n");
   puts("    D2F <A or B> <filename> -D\n");
   puts("If <filename> exists, you will be asked to confirm the overwrite.\n\n");
   puts("NOTE: Currently, this program handles 160K, 180K, 320K, 360K & 1.2M 5.25\" disks,\n");
   puts("      and 720K & 1.44M 3.5\" disks.\n\n");
   puts("      A screenful of user comments can be added, and the disk directory\n");
   puts("      tree (with expansion of ZIP/ARJ/ARC files) will be included in the\n");
   puts("      saved disk image.\n\n");
   puts("If the -D flag is given, the image name, comments, and directory list\n");
   puts("will be NOT appended to the file C:\\DISKLIST.DB\n");
   exit(1);
   }

ident_disk()
   {
   WORD numcyls, flag;
   WORD numheads;

   flag = FALSE;  /* do not have BIOS params */
/*
 *  NOTE: I'm going to assume that we HAVE to be able to read
 *        the first sector, even if it doesn't contain a BPB.
 *
 *        Otherwise, assume that there is no disk in the drive.
 */
   if (!phys_sector_io(diskbuf,DISK_READ,1,0,0,1,drive))
      {
      puts("Error:  no disk in drive, or bad disk in drive.\n");
      return(FALSE);
      }
          /* check for valid BPB */
/*
printf("BPS=%04x RS=%d NFATs=%d HS=%d NHeads=%d\n",
*(WORD *)&diskbuf[0x0b],
*(WORD *)&diskbuf[0x0e],
diskbuf[0x10],
*(WORD *)&diskbuf[0x1c],
*(WORD *)&diskbuf[0x1a]);
ci();
*/
   if ((*(WORD *)&diskbuf[0x0b] == SECTOR_SIZE)  /* bps */
         && (*(WORD *)&diskbuf[0x0e] == 1)  /* Reserved sectors */
         && (diskbuf[0x10] == 2)  /* Nfats */
         && (*(WORD *)&diskbuf[0x1c] == 0)  /* hidden sectors */
         && ((*(WORD *)&diskbuf[0x1a] == 1)  /* Num Heads is 1 */
                    || (*(WORD *)&diskbuf[0x1a] == 2)))  /* or 2 */
      {
      numheads = *(WORD *)&diskbuf[0x1a];
      numcyls = *(WORD *)&diskbuf[0x13] / numheads;  /* total sectors / heads */
      numcyls /= *(WORD *)&diskbuf[0x18];  /* / SPT = cyls */
/*      if ((*(WORD *)&diskbuf[0x1a] != MAX_HEADS) || */
      if (((*(WORD *)&diskbuf[0x18] != SPT_144)
                            && (*(WORD *)&diskbuf[0x18] != SPT_12)
                            && (*(WORD *)&diskbuf[0x18] != SPT_720360)
                            && (*(WORD *)&diskbuf[0x18] != SPT_320160))
             || ((numcyls != MAX_CYL) && (numcyls != MAX_CYL_360)))
         {
/*
printf("numcyls = %d, numheads = %d\n",numcyls,numheads);
*/
         puts("Error, this disk is not a standard 160K/180K/320K/360K/720K/1.2M/1.44M floppy\n");
         return(FALSE);
         }
      else
         {
         flag = TRUE;
         cyls = numcyls;
         heads = numheads;
         spt = *(WORD *)&diskbuf[0x18];
         }
      }
   if (!flag)  /* can't get params from BPB */
      {
      if (phys_sector_io(diskbuf,DISK_READ,1,0,0,SPT_144+1,drive))
         {
         puts("Error, this program does not work with disks above 1.44M 3.5\" disks.\n");
         return(FALSE);
         }
      if (phys_sector_io(diskbuf,DISK_READ,1,MAX_HEADS-1,MAX_CYL-1,SPT_144,drive))
         {
         cyls = MAX_CYL;
         heads = MAX_HEADS;
         spt = SPT_144;
         }
      else
         {
         if (phys_sector_io(diskbuf,DISK_READ,1,MAX_HEADS-1,MAX_CYL-1,SPT_12,drive))
            {
            cyls = MAX_CYL;
            heads = MAX_HEADS;
            spt = SPT_12;
            }
         else
            {
            if (phys_sector_io(diskbuf,DISK_READ,1,MAX_HEADS-1,MAX_CYL-1,SPT_720360,drive))
               {
               cyls = MAX_CYL;
               heads = MAX_HEADS;
               spt = SPT_720360;
               }
            else
               {
               if (phys_sector_io(diskbuf,DISK_READ,1,MAX_HEADS-1,MAX_CYL_360-1,SPT_720360,drive))
                  {
                  cyls = MAX_CYL_360;
                  heads = MAX_HEADS;
                  spt = SPT_720360;
                  }
               else
                  {
                  if (phys_sector_io(diskbuf,DISK_READ,1,MAX_HEADS-1,MAX_CYL_360-1,SPT_320160,drive))
                     {
                     cyls = MAX_CYL_360;
                     heads = MAX_HEADS;
                     spt = SPT_320160;
                     }
                  else
                     {
                     if (phys_sector_io(diskbuf,DISK_READ,1,0,MAX_CYL_360-1,SPT_720360,drive))
                        {
                        cyls = MAX_CYL_360;
                        heads = 1;
                        spt = SPT_720360;
                        }
                     else
                        {
                        if (phys_sector_io(diskbuf,DISK_READ,1,0,MAX_CYL_360-1,SPT_320160,drive))
                           {
                           cyls = MAX_CYL_360;
                           heads = 1;
                           spt = SPT_320160;
                           }
                        else
                           {
                           puts("Error, this program only works with 160K/180K/320K/360K/720K/1.2M/1.44M disks\n");
                           return(FALSE);
                           }  /* 160K */
                        }  /* 180K */
                     } /* 320K */
                  }  /* 360K */
               }  /* 720K */
            }  /* 1.2M */
         }  /* 1.44M */
      }  /* (!flag) */
   strcpy(header.signature,sig);
   header.version = VERSION;
   header.sectorcount = spt;
   header.cylcount = cyls;
   header.headcount = heads;
   header.head0extra1 = FALSE;
   header.head1extra1 = FALSE;
   header.head0extra2 = FALSE;
   header.head1extra2 = FALSE;
   if (do_extra_tracks)
      {
      if (phys_sector_io(diskbuf,DISK_READ,1,0,cyls,1,drive))
         header.head0extra1 = TRUE;
      if (phys_sector_io(diskbuf,DISK_READ,1,1,cyls,1,drive))
         header.head1extra1 = TRUE;
      if (phys_sector_io(diskbuf,DISK_READ,1,0,cyls+1,1,drive))
         header.head0extra2 = TRUE;
      if (phys_sector_io(diskbuf,DISK_READ,1,1,cyls+1,1,drive))
         header.head1extra2 = TRUE;
      }
   return(TRUE);
   }

build_outbuf(h,c)
   WORD h, c;
   {
   WORD s;
   BYTE *p, *p2;

   printf("H: %02x  C: %02x\r",h,c);
   outbuf_count = 0;
   if (phys_sector_io(diskbuf,DISK_READ,spt,h,c,1,drive))
      {
      for (s = 1; s <= spt; s++)
         set_outbuf_hdr(&diskbuf[(s-1)*SECTOR_SIZE],h,c,s);
      }
   else
      {
      for (s = 1; s <= spt; s++)
         {
         if (phys_sector_io(diskbuf,DISK_READ,1,h,c,s,drive))
            set_outbuf_hdr(diskbuf,h,c,s);
         else
            {
            printf("\n          Disk read error  H: %02x  C: %02x  S: %02x\n",h,c,s);
            p = outbuf;
            p += outbuf_count;
            *(p++) = SFLAG_BAD;
            *(p++) = h;  /* set head */
            *(p++) = s;  /* set sector */
            *(p++) = c;  /* set cyl */
            *(WORD *)p = 0;  /* clear CRC value */
            outbuf_count += 6;
            }
         }
      }
   }

set_outbuf_hdr(buf,h,c,s)
   BYTE *buf;
   WORD h, c, s;
   {
   WORD i,val,flag,crc,b;
   BYTE *p, *p2;

   p = outbuf;
   p += outbuf_count;
   p2 = p;  /* save this pointer for a moment */
   p++;   /* skip flag byte */
   *(p++) = h;  /* set head */
   *(p++) = s;  /* set sector */
   *(p++) = c;  /* set cyl */
   flag = TRUE;
   val = buf[0];
   crc = 0;
   for (i = 0; i < SECTOR_SIZE; i++)
      {
      b = buf[i];
      crc = updatecrc(crc,b);
      if (val != b)
         flag = FALSE;
      *(p+2+i) = b;
      }
   if (flag)  /* all bytes the same */
      {
      *p2 = SFLAG_REPEAT;
      *(p++) = val;
      *(p++) = 0;
      outbuf_count += 6;
      }
   else
      {
      *p2 = SFLAG_NORMAL;
      *(WORD *)p = crc;
      p += 2 + SECTOR_SIZE;
      outbuf_count += 6 + SECTOR_SIZE;
      }
   }

write_outbuf()
   {
   if (write(handle,outbuf,outbuf_count) != outbuf_count)
      {
      close(handle);
      unlink(filename);
      puts("\nError writing to output file.  Disk full/write protected?\n");
      exit(2);
      }
   }

/* update crc */
WORD updatecrc(crc,c)
WORD crc, c;
   {
   WORD tmp;

   tmp=crc^c;
   crc=(crc>>8)^crc16tab[tmp & 0xff];
   return(crc);
   }

phys_sector_io(buf,function,count,head,cyl,sector,drive)
             /* +4    +6      +8  +10  +12  +14    +16 */
   BYTE *buf;
   WORD function,head,cyl,sector,drive;
   {
#asm
   mov  di,3                    ;retry count
physsectio_retry:
   mov  cx,word [bp+12]         ;get cyl
   ror  ch,1                    ;low 2 bits
   ror  ch,1                    ;to high 2 bits
   xchg cl,ch                   ;put in right order
   or   cl,byte [bp+14]         ;add sector
   mov  dl,byte [bp+16]         ;get drive
   mov  dh,byte [bp+10]         ;get head
   push ds                      ;data seg
   pop  es                      ;to es
   mov  bx,word [bp+4]          ;es:bx = ptr to buffer
   mov  ah,byte [bp+6]          ;get function
   mov  al,byte [bp+8]          ;sector count
   int  13h                     ;call BIOS
   mov  ax,1                    ;default to OK
   jnc  physsectio_done         ;NC, we're done
   xor  ax,ax                   ;fn = reset disks
   int  13h                     ;call BIOS
   dec  di                      ;count this try
   jnz  physsectio_retry        ;not done, try again
   xor  ax,ax                   ;set error
physsectio_done:
#endasm
   }

get_comments()
   {
   WORD r,c,i;
   WORD ch, mode;

   cls();
   atsay(MAX_COMMENT_LINES,0,"Enter comment.  <cursor keys> to move, Esc to exit");
/*
   for (i = 0; i < MAX_COMMENT_LINES; i++)
      atchr(i,0,'>');
*/
   r = 0;
   c = 0;
   mode = MODE_INSERT;
   atsay(MAX_COMMENT_LINES,66,"Insert    ");
   do
      {
      ch = gchar(r,c);
      switch (ch)
         {
         case CR:
            if (r < MAX_COMMENT_LINES-1)
               {
               r++;
               c = 0;
               }
            break;
         case BS:
            if (c)
               {
               c--;
               for (i = c; i < comments.line[r].linelen; i++)
                  comments.line[r].buf[i] = comments.line[r].buf[i+1];
               comments.line[r].linelen--;
               atsay(r,0,comments.line[r].buf);
               atchr(r,0+comments.line[r].linelen,' ');
               }
            else
               if (r)
                  {
                  r--;
                  c = comments.line[r].linelen;
                  }
            break;
         case DEL:
            if (c < comments.line[r].linelen)
               {
               for (i = c; i < comments.line[r].linelen; i++)
                  comments.line[r].buf[i] = comments.line[r].buf[i+1];
               comments.line[r].linelen--;
               atsay(r,0,comments.line[r].buf);
               atchr(r,0+comments.line[r].linelen,' ');
               }
            break;
         case UPARROW:
            if (r)
               {
               r--;
               c = MIN(c,comments.line[r].linelen);
               }
            else
               c = 0;
            break;
         case PGUPKEY:
            r = 0;
            c = 0;
            break;
         case HOMEKEY:
            c = 0;
            break;
         case LEFTARROW:
            if (c)
               c--;
            else
               if (r)
                  {
                  r--;
                  c = comments.line[r].linelen;
                  }
            break;
         case RIGHTARROW:
            if (c < comments.line[r].linelen)
               c++;
            else
               if (r < MAX_COMMENT_LINES-1)
                  {
                  r++;
                  c = 0;
                  }
            break;
         case ENDKEY:
            c = comments.line[r].linelen;
            break;
         case DOWNARROW:
            if (r < MAX_COMMENT_LINES-1)
               {
               r++;
               c = MIN(c,comments.line[r].linelen);
               }
            else
               c = comments.line[r].linelen;
            break;
         case PGDNKEY:
            r = MAX_COMMENT_LINES-1;
            c = comments.line[r].linelen;
            break;
         case INSERT:
            if (mode == MODE_INSERT)
               {
               mode = MODE_OVERSTRIKE;
               atsay(MAX_COMMENT_LINES,66,"Overstrike");
               }
            else
               {
               mode = MODE_INSERT;
               atsay(MAX_COMMENT_LINES,66,"Insert    ");
               }
            break;
         default:
            if ((ch >= SPACE) && (ch <= MAXCHAR))
               insert_comment_ch(&r,&c,ch,mode);
            break;
         }
      } while (ch != ESC_KEY);
   atsay(MAX_COMMENT_LINES,0,eraseline);
   curpos(MAX_COMMENT_LINES,0);
   }

insert_comment_ch(r,c,ch,mode)
   WORD *r, *c, ch, mode;
   {
   WORD i;

   if ((comments.line[*r].linelen) >= MAX_COMMENT_LINE_LEN)
      {
      if (*r < MAX_COMMENT_LINES)
         {
         (*r)++;
         *c = 0;
         }
      }
   if ((*r < MAX_COMMENT_LINES)
            && ((comments.line[*r].linelen) < MAX_COMMENT_LINE_LEN))
      {
      if ((mode == MODE_INSERT) && (*c < comments.line[*r].linelen))
         {
         for (i = comments.line[*r].linelen; i >= *c; i--)
            comments.line[*r].buf[i+1] = comments.line[*r].buf[i];
         }
      comments.line[*r].buf[*c] = ch;
      if ((mode == MODE_INSERT) || (*c == comments.line[*r].linelen))
         comments.line[*r].linelen++;
      (*c)++;
      atsay(*r,0,eraseline);
      atsay(*r,0,comments.line[*r].buf);
      }
   }

add_comments()
   {
   WORD i, j, len;
   BYTE *p;

   *(DWORD *)diskbuf = SFLAG_COMMENT;
   p = &diskbuf[sizeof(struct sect_hdr)];
   *p = '\0';  /* init string */
   len = 0;
   for (i = 0; i < MAX_COMMENT_LINES; i++)
      {
      if (comments.line[i].linelen)
         {
         strcat(p,comments.line[i].buf);
         strcat(p,"\015\012");
         len += comments.line[i].linelen + 2;
         for (j = i+1; j < MAX_COMMENT_LINES; j++)
            if (comments.line[j].linelen)  /* not done yet! */
               {
               while (--j > i)
                  {
                  strcat(p,"\015\012");
                  len += 2;
                  }
               break;  /* exit FOR loop! */
               }
         }
      }
   if (len)
      {
      *(WORD *)&diskbuf[4] = len;
      if (write(handle,diskbuf,sizeof(struct sect_hdr)+len)
                != sizeof(struct sect_hdr)+len)
         return(FALSE);
      if (update_database)
         if ((write(db_handle,&diskbuf[sizeof(struct sect_hdr)],len) != len)
                    || (write(db_handle,db_field_str,3) != 3))
            return(FALSE);
      }
   return(TRUE);
   }

cls()
   {
#asm
   mov  ah,0fh            ;fn = get video mode
   int  10h               ;call BIOS
   mov  ah,0              ;fn = set video mode (al = mode from above)
   int  10h               ;call BIOS
#endasm
   }

atchr(row,col,ch)
   WORD row, col;
   BYTE ch;
   {
   BYTE str[2];

   str[0] = ch;
   str[1] = '\0';
   curpos(row,col);
   puts(str);
   atsay(row,col,str);
   }

atsay(row,col,str)
   WORD row, col;
   BYTE *str;
   {
   curpos(row,col);
   puts(str);
   }

curpos(row,col)
   WORD row, col;
   {
#asm
   mov  dh,byte [bp+4]    ;get row
   mov  dl,byte [bp+6]    ;get col
   mov  bh,0              ;video page
   mov  ah,2              ;fn = set cursor pos
   int  10h               ;call BIOS
#endasm
   }

gchar(row,col)
   WORD row, col;
   {
   WORD ch;

   curpos(row,col);
   if ((ch = ci()) == 0)
      ch = ci() + CURSOR_FLAG;
   return(ch);
   }

add_dirlist()
   {
#asm
   mov ah,19h                 ;get current drive
   int 21h                    ;call DOS
   inc al                     ;1=A, 2=B, etc
   mov byte saved_drive_,al   ;save for later cleanup
   mov dl,byte drive_         ;get desired drive
   mov ah,0eh                 ;fn = select disk
   int 21h                    ;call DOS
   dec al                     ;make A=0, B=1, C=2, etc
   cmp al,byte drive_         ;can this drive exist?
   jae good_drive             ;yes, keep it
   mov dl,byte saved_drive_   ;else get saved drive
   dec dl                     ;change to A=0, B=1, etc
   mov ah,0eh                 ;fn = select disk
   int 21h                    ;call DOS
   mov byte saved_drive_,0ffh ;and set to error on saved_drive
good_drive:
#endasm
   if (saved_drive == '\377')
      return(FALSE);
   strcpy(currentdir[0],"\\");
   getdir(dirname);   /* get current dir */
   strcat(currentdir[0],dirname);
   strcpy(dirname,currentdir[0]);
   currentdir[dirlevel][0] = '\0';
   chdir("\\");
   start = 0;
   do
      {
      get_dirlist();
      } while (nextdir());
#asm
   mov dl,byte saved_drive_   ;get our original drive
   dec dl                     ;make A=0, B=1, etc
   mov ah,0eh                 ;fn = select drive
   int 21h                    ;call DOS
#endasm
   }

int nextdir()
   {
   int i;

   do
      {
      if (start == 0)
         {
         i = find_first(wildname,0x17,fname,dta[dirlevel]);
         start = 1;
         }
      else
         i = find_next(wildname,0x17,fname,dta[dirlevel]);
      } while ((i == 0) && (((dta[dirlevel][21] & 0x10) == 0) || (strcmp(fname,".") == 0)
                           || (strcmp(fname,"..") == 0)));
   if ((i == 0) && ((dta[dirlevel][21] & 0x10) == 0x10) && (strcmp(fname,"."))
                         && (strcmp(fname,"..")) && (dirlevel < DEPTH-1))
      {
      dirlevel++;
      if (dirlevel > max_dirlevel)
         max_dirlevel = dirlevel;
      start = 0;
      strcpy(currentdir[dirlevel],fname);
      chdir(fname);
      }
   else
      if ((i) || (dirlevel >= DEPTH-1))
         {
         dirlevel--;
         chdir("..");
         if (dirlevel)
            return(nextdir());
         }
   return(dirlevel);
   }

void get_dirlist()
   {
   WORD y,m,d;
   WORD hh,mm,ss;
   BYTE ampm;
   WORD i;
   DWORD zip_loc,arc_loc,arj_loc;
   int filein;
   BYTE buf[256];  /* big for ARJ use */
   BYTE *p;

   path[0] = '\0';
   diskbuf[0] = '\0';
   for (i = 1; i <= dirlevel; i++)
      {
      strcat(path,currentdir[i]);
      if (i > 1)
         {
         strcat(path,"\\");
         }
      }
   if (find_first(wildname,0,fname,dta[0]) == 0)
      {
      do
         {
         y = (((dta[0][25] >> 1) & 0x7f) + 80) % 100;
         m = 8 * (dta[0][25] & 1) + ((dta[0][24] >> 5) & 7);
         d = dta[0][24] & 0x1f;
         hh = (dta[0][23] >> 3) & 0x1f;
         if (hh > 12)
            {
            ampm = 'p';
            hh -= 12;
            }
         else
            ampm = 'a';
         mm = ((dta[0][22] >> 5) & 0x07) + ((dta[0][23] & 0x07) << 3);
         ss = (dta[0][22] & 0x1f) * 2;
         strcpy(outbuf,path);
         sprintf(&outbuf[strlen(outbuf)],"%-12s   %7ld  %2d-%02d-%02d  %2d:%02d:%02d%c\015\012",
                     fname,*((long *)&dta[0][26]),m,d,y,hh,mm,ss,ampm);
/*
puts(outbuf);
*/
         if (strlen(diskbuf)+strlen(outbuf) > DISKBUF_SIZE-2)
            {
            header.extrasflag |= EFLAG_DIR;
            *(DWORD *)buf = SFLAG_DIR;
            *(WORD *)&buf[4] = strlen(diskbuf);
            if (write(handle,buf,6) != 6)
               return(FALSE);
            if (write(handle,diskbuf,strlen(diskbuf)) != strlen(diskbuf))
               return(FALSE);
            if (update_database)
               if (write(db_handle,diskbuf,strlen(diskbuf)) != strlen(diskbuf))
                  return(FALSE);
            diskbuf[0] = '\0';
            }
         strcat(diskbuf,outbuf);
         if (strcmp(&fname[strlen(fname)-4],".ZIP") == 0)
            {
            if ((filein = open(fname,0)) != -1)
               {
               while (TRUE)  /* use BREAK to get out */
                  {
                  zip_loc = lseek(filein,0L,1);  /* get current pos */
                  if (read(filein,&zip_localhdr,sizeof(zip_localhdr)) < 0x1f)
                     break;  /* too small! */
                  if (zip_localhdr.signature == ZIP_CENTRALHDR)
                     break;  /* at end of .ZIP file */
                  if (zip_localhdr.signature != ZIP_LOCALHDR)
                     break;  /* invalid .ZIP file */
                  zip_localhdr.filename[zip_localhdr.namelen] = '\0';  /* set EOS */
                  m = (zip_localhdr.file_date >>5) & 0x0f;
                  d = zip_localhdr.file_date & 0x1f;
                  y = (((zip_localhdr.file_date >> 9) & 0x7f) + 80) % 100;
                  hh = (zip_localhdr.file_time >> 11) & 0x1f;
                  if (hh > 12)
                     {
                     ampm = 'p';
                     hh -= 12;
                     }
                  else
                     ampm = 'a';
                  mm = (zip_localhdr.file_time >> 5) & 0x3f;
                  ss = (zip_localhdr.file_time & 0x1f) * 2;
                  sprintf(outbuf,"      %-12s   %7ld  %2d-%02d-%02d  %2d:%02d:%02d%c\015\012",
                            zip_localhdr.filename,zip_localhdr.realsize,m,d,y,hh,mm,ss,ampm);
/*
puts(outbuf);
*/
                  if (strlen(diskbuf)+strlen(outbuf) > DISKBUF_SIZE-2)
                     {
                     header.extrasflag |= EFLAG_DIR;
                     *(DWORD *)buf = SFLAG_DIR;
                     *(WORD *)&buf[4] = strlen(diskbuf);
                     if (write(handle,buf,6) != 6)
                        return(FALSE);
                     if (write(handle,diskbuf,strlen(diskbuf)) != strlen(diskbuf))
                        return(FALSE);
                     if (update_database)
                        if (write(db_handle,diskbuf,strlen(diskbuf)) != strlen(diskbuf))
                           return(FALSE);
                     diskbuf[0] = '\0';
                     }
                  strcat(diskbuf,outbuf);
                  if (!zip_localhdr.comprsize)  /* if size in data descriptor */
                     break;  /* can't find next header */
                  zip_loc += ZIP_HDRSIZE + zip_localhdr.comprsize
                                         + zip_localhdr.namelen
                                         + zip_localhdr.extra_field_len;
                  if (zip_localhdr.bitflag & ZIP_HAVE_DATA_DESC)
                     zip_loc += 12;  /* 3 extra DWORDs */
                  lseek(filein,zip_loc,0);
                  }
               close(filein);
               }
            }  /* if ZIP */
         if (strcmp(&fname[strlen(fname)-4],".ARC") == 0)
            {
            if ((filein = open(fname,0)) != -1)
               {
               while (TRUE)  /* use BREAK to get out */
                  {
                  arc_loc = lseek(filein,0L,1);  /* get current pos */
                  if (read(filein,&arc_header,sizeof(arc_header)) != sizeof(arc_header))
                     break;  /* too small! */
                  if (arc_header.signature != ARC_SIGNATURE)
                     break;  /* invalid .ARC file */
                  if (arc_header.stowage == ARC_EOF)
                     break;  /* end of ARC */
                  m = (arc_header.file_date >>5) & 0x0f;
                  d = arc_header.file_date & 0x1f;
                  y = (((arc_header.file_date >> 9) & 0x7f) + 80) % 100;
                  hh = (arc_header.file_time >> 11) & 0x1f;
                  if (hh > 12)
                     {
                     ampm = 'p';
                     hh -= 12;
                     }
                  else
                     ampm = 'a';
                  mm = (arc_header.file_time >> 5) & 0x3f;
                  ss = (arc_header.file_time & 0x1f) * 2;
                  sprintf(outbuf,"      %-12s   %7ld  %2d-%02d-%02d  %2d:%02d:%02d%c\015\012",
                            arc_header.filename,arc_header.realsize,m,d,y,hh,mm,ss,ampm);
/*
puts(outbuf);
*/
                  if (strlen(diskbuf)+strlen(outbuf) > DISKBUF_SIZE-2)
                     {
                     header.extrasflag |= EFLAG_DIR;
                     *(DWORD *)buf = SFLAG_DIR;
                     *(WORD *)&buf[4] = strlen(diskbuf);
                     if (write(handle,buf,6) != 6)
                        return(FALSE);
                     if (write(handle,diskbuf,strlen(diskbuf)) != strlen(diskbuf))
                        return(FALSE);
                     if (update_database)
                        if (write(db_handle,diskbuf,strlen(diskbuf)) != strlen(diskbuf))
                           return(FALSE);
                     diskbuf[0] = '\0';
                     }
                  strcat(diskbuf,outbuf);
                  arc_loc += ARC_HDRSIZE + arc_header.comprsize;
                  lseek(filein,arc_loc,0);
                  }
               close(filein);
               }
            }  /* if ARC */
         if (strcmp(&fname[strlen(fname)-4],".ARJ") == 0)
            {
            if ((filein = open(fname,0)) != -1)
               {
               read(filein,&arj_header,4);
               arj_loc = 4 + arj_header.basic_header_size + 4;
               lseek(filein,arj_loc,0);
               read(filein,&i,2);
               arj_loc += 2;
               if (i)
                  arj_loc += i + 4;
               lseek(filein,arj_loc,0);
               while (TRUE)  /* use BREAK to get out */
                  {
                  if (!parse_arj_header(filein,&arj_loc,buf))
                     break;
/* printf("2: filename found = %s  size=%ld\n",buf,arj_header.realsize); */
                  m = (arj_header.file_date >>5) & 0x0f;
                  d = arj_header.file_date & 0x1f;
                  y = (((arj_header.file_date >> 9) & 0x7f) + 80) % 100;
                  hh = (arj_header.file_time >> 11) & 0x1f;
                  if (hh > 12)
                     {
                     ampm = 'p';
                     hh -= 12;
                     }
                  else
                     ampm = 'a';
                  mm = (arj_header.file_time >> 5) & 0x3f;
                  ss = (arj_header.file_time & 0x1f) * 2;
/* printf("3: filename found = %s  size=%ld\n",buf,arj_header.realsize); */
                  sprintf(outbuf,"      %-12s   %7ld  %2d-%02d-%02d  %2d:%02d:%02d%c\015\012",
                         buf,arj_header.realsize,m,d,y,hh,mm,ss,ampm);
/*
puts(outbuf);
*/
                  if (strlen(diskbuf)+strlen(outbuf) > DISKBUF_SIZE-2)
                     {
                     header.extrasflag |= EFLAG_DIR;
                     *(DWORD *)buf = SFLAG_DIR;
                     *(WORD *)&buf[4] = strlen(diskbuf);
                     if (write(handle,buf,6) != 6)
                        return(FALSE);
                     if (write(handle,diskbuf,strlen(diskbuf)) != strlen(diskbuf))
                        return(FALSE);
                     if (update_database)
                        if (write(db_handle,diskbuf,strlen(diskbuf)) != strlen(diskbuf))
                           return(FALSE);
                     diskbuf[0] = '\0';
                     }
                  strcat(diskbuf,outbuf);
                  }  /* end WHILE */
               close(filein);
               }
            }  /* if ARJ */
         } while (find_next(wildname,0,fname,dta[0]) == 0);
      }
   if (strlen(diskbuf))
      {
      header.extrasflag |= EFLAG_DIR;
      *(DWORD *)buf = SFLAG_DIR;
      *(WORD *)&buf[4] = strlen(diskbuf);
      if (write(handle,buf,6) != 6)
         return(FALSE);
      if (write(handle,diskbuf,strlen(diskbuf)) != strlen(diskbuf))
         return(FALSE);
      if (update_database)
         if (write(db_handle,diskbuf,strlen(diskbuf)) != strlen(diskbuf))
            return(FALSE);
      }
   return(TRUE);
   }

parse_arj_header(filein,arj_loc,buf)
   int filein;
   DWORD *arj_loc;
   BYTE *buf;
   {
   WORD i,len;
WORD j;
BYTE *s, *d, *t, *c;

/*
printf("0: *arj_loc = %04lx, filepos() = %04lx\n",*arj_loc,lseek(filein,0L,1));
*/
   if ((read(filein,&arj_header,sizeof(arj_header)) != sizeof(arj_header))
             || (arj_header.signature != ARJ_SIGNATURE)
             || (arj_header.basic_header_size > ARJ_MAX_HEADER_SIZE))
      return(FALSE);
   *arj_loc = lseek(filein,0L,1) - sizeof(arj_header) + 4;  /* get base pos */
/*
s = &arj_header.signature;
d = &arj_header.file_date;
t = &arj_header.file_time;
c = &arj_header.comprsize;

printf("1: arj_loc=%04lx  sizeof(arj_header) = %d\n",*arj_loc,sizeof(arj_header));
printf("   arj_header.signature         = %04x (%04x)\n",arj_header.signature,s);
printf("   arj_header.basic_header_size = %04x\n",arj_header.basic_header_size);
printf("   arj_header.first_header_size = %04x\n",arj_header.first_header_size);
printf("   arj_header.archiver_version  = %04x\n",arj_header.archiver_version);
printf("   arj_header.min_xtract_vers   = %04x\n",arj_header.min_extract_version);
printf("   arj_header.host_os           = %04x\n",arj_header.host_os);
printf("   arj_header.arj_flags         = %04x\n",arj_header.arj_flags);
printf("   arj_header.method            = %04x\n",arj_header.method);
printf("   arj_header.file_type         = %04x\n",arj_header.file_type);
printf("   arj_header.reserved          = %04x\n",arj_header.reserved);

printf("   arj_header.file_date         = %04x (%04x, %d)\n",arj_header.file_date,
                            d,d-s);
printf("   arj_header.file_time         = %04x (%04x, %d)\n",arj_header.file_time,
                            t,t-s);

printf("   arj_header.comprsize         = %04lx (%04x, %d)\n",arj_header.comprsize,
                            c,c-s);
printf("   arj_header.realsize          = %04lx\n",arj_header.realsize);
printf("   arj_header.crc32             = %04lx\n",arj_header.crc32);
printf("   arj_header.filespec_pos      = %04x\n",arj_header.filespec_pos);
printf("   arj_header.fileaccess_mode   = %04x\n",arj_header.fileaccess_mode);
printf("   arj_header.host_data         = %04x\n",arj_header.host_data);
*/
   *arj_loc += arj_header.first_header_size;
/*
printf("2: arj_loc=%04lx\n",*arj_loc);
*/
   if (arj_header.arj_flags & ARJ_EXTFILE_FLAG)
      *arj_loc + ARJ_EXTFILE_SIZE;
/*
printf("3: arj_loc=%04lx\n",*arj_loc);
*/
   lseek(filein,*arj_loc,0);  /* lseek BOF */
/*
printf("*arj_loc = %04lx, filepos() = %04lx\n",*arj_loc,lseek(filein,0L,1));
*/
   if ((len = read(filein,outbuf,ARJ_MAX_HEADER_SIZE)) < 1+1+ARJ_CRC_SIZE+2)
      return(FALSE);
/*
printf("len = %d\n",len);
for (j = 0; j < 2; j++)
   {
   for (i = 0; i < 16; i++)
      printf("%02x ",outbuf[j*16+i]);
   printf("\n");
   }
*/
   if ((i = strlen(outbuf) + 1) > len)
      return(FALSE);
   if (i < arj_header.filespec_pos)
      return(FALSE);
   if (strlen(&outbuf[arj_header.filespec_pos]) > 120)  /* too big! */
      return(FALSE);
   strcpy(buf,&outbuf[arj_header.filespec_pos]);
/* printf("1: filename found = %s  size=%ld\n",buf,arj_header.realsize); */
   if (len < i+1+ARJ_CRC_SIZE+2)  /* not enough for comment EOS + CRC + 1st ext. hdr size */
      return(FALSE);
/*
printf("4: arj_loc=%04lx\n",*arj_loc);
*/
   *arj_loc += i + strlen(&outbuf[i]) + 1 + ARJ_CRC_SIZE + 2;
/*
printf("5: arj_loc=%04lx\n",*arj_loc);
*/
   i += strlen(&outbuf[i]) + 1 + ARJ_CRC_SIZE;  /* ofs to 1st extended header size */
   if (len < i+2)  /* not enough read for extended header size */
      return(FALSE);
/*
printf("5a: ext_hdr_size=%d, i=%d\n",*(WORD *)&outbuf[i],i);
for (j = 0; j < i+2; j++)
   {
   printf("%02x ",outbuf[j]);
   if (((j+1)&0x0f) == 0x0f)
      printf("\n");
   }
printf("\n");
*/
   if (*(WORD *)&outbuf[i])  /* if extended header */
      *arj_loc += *(WORD *)&outbuf[i] + ARJ_CRC_SIZE;
/*
printf("6: arj_loc=%04lx\n",*arj_loc);
*/
   *arj_loc += arj_header.comprsize;
   lseek(filein,*arj_loc,0);  /* goto next header */
/*
printf("7: arj_loc = %04lx  filepos()=%04lx\n",*arj_loc,lseek(filein,0L,1));
*/
   return(TRUE);
   }


int find_first(wildname,attrib,fname,dta_addr)
   char *wildname, *fname, *dta_addr;
   int attrib;
   {
   int return_val;

   return_val = 0;
#asm
   mov ah,2fh            ;fn = get DTA
   int 21h               ;call DOS
   mov word dta_ofs_,bx  ;save old DTA offset
   mov word dta_seg_,es  ;save old DTA segment
   push ds               ;copy DS
   pop es                ;to ES
   mov dx,word [bp+10]   ;DS:DX = new DTA
   mov ah,1ah            ;fn = set DTA
   int 21h               ;call DOS
   mov dx,word [bp+4]    ;DS:DX = wild card fname
   mov cx,word [bp+6]    ;attribute
   mov ah,4eh            ;fn = find first
   int 21h               ;call DOS
   jae f_f_1             ;JNC f_f_1
   mov word [bp-2],ax    ;else save error
f_f_1:
   mov di,word [bp+8]    ;ES:DI = target string
   mov si,word [bp+10]   ;DS:SI = this DTA
   add si,30             ;DS:SI = source string
   cld                   ;UP!
f_f_2:
   lodsb                 ;get byte
   stosb                 ;store byte
   cmp al,0              ;EOS?
   jne f_f_2             ;no, continue
   mov dx,word dta_ofs_  ;get old DTA offset
   mov ds,word dta_seg_  ;get old DTA segment
   mov ah,1ah            ;fn = set DTA
   int 21h               ;call DOS
   push es               ;copy ES
   pop ds                ;to DS
#endasm
   return(return_val);
   }

int find_next(wildname,attrib,fname,dta_addr)
   char *wildname, *fname, *dta_addr;
   int attrib;
   {
   int return_val;

   return_val = 0;
#asm
   mov ah,2fh            ;fn = get DTA
   int 21h               ;call DOS
   mov word dta_ofs_,bx  ;save old DTA offset
   mov word dta_seg_,es  ;save old DTA segment
   push ds               ;copy DS
   pop es                ;to ES
   mov dx,word [bp+10]   ;DS:DX = new DTA
   mov ah,1ah            ;fn = set DTA
   int 21h               ;call DOS
   mov ah,4fh            ;fn = find next
   int 21h               ;call DOS
   jae f_n_1             ;JNC f_n_1
   mov word [bp-2],ax    ;else save error
f_n_1:
   mov di,word [bp+8]    ;ES:DI = target string
   mov si,word [bp+10]   ;DS:SI = this DTA
   add si,30             ;DS:SI = source string
   cld                   ;UP!
f_n_2:
   lodsb                 ;get byte
   stosb                 ;store byte
   cmp al,0              ;EOS?
   jne f_n_2             ;no, continue
   mov dx,word dta_ofs_  ;get old DTA offset
   mov ds,word dta_seg_  ;get old DTA segment
   mov ah,1ah            ;fn = set DTA
   int 21h               ;call DOS
   push es               ;copy ES
   pop ds                ;to DS
#endasm
   return(return_val);
   }

void getdir(str)
   char *str;
   {
#asm
   mov dl,0                   ;default drive
   mov si,word [bp+4]         ;get destination pointer
   mov ah,47h                 ;fn = GET DIR
   int 21h                    ;call DOS
#endasm
   }

void chdir(str)
   char *str;
   {
#asm
   mov dx,word [bp+4]    ;DS:DX = pathname
   mov ah,3bh            ;fn = CHDIR
   int 21h               ;call DOS
#endasm
   }

truncate(name,pos)
   BYTE *name;
   DWORD pos;
   {
#asm
   mov  dx,word [bp+4]     ;ds:dx = name
   mov  ax,3d02h           ;fn = open for read/write
   int  21h                ;call DOS
   jc   truncate_exit      ;nope, quit!
   mov  bx,ax              ;handle to bx
   mov  cx,word [bp+8]     ;cx:dx = location in file
   mov  dx,word [bp+6]
   mov  ax,4200h           ;fn = lseek BOF+CX:DX
   int  21h
   mov  dx,word [bp+4]     ;ds:dx = don't care pointer
   mov  cx,0               ;size to write = 0 (truncate file)
   mov  ah,40h             ;fn = write to file
   int  21h                ;call DOS
   mov  ah,3eh             ;fn = close file
   int  21h                ;call DOS
truncate_exit:
#endasm
   }
